@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2011 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2012 -- leonerd@leonerd.org.uk
 
 package IO::Socket::IP;
 
@@ -9,17 +9,17 @@ use strict;
 use warnings;
 use base qw( IO::Socket );
 
-our $VERSION = '0.07_007';
+our $VERSION = '0.08_001';
 
 use Carp;
 
-use Socket 1.95 qw(
+use Socket 1.97 qw(
    getaddrinfo getnameinfo
    AF_INET
    AI_PASSIVE
    IPPROTO_TCP IPPROTO_UDP
-   IPV6_V6ONLY
-   NI_DGRAM NI_NUMERICHOST NI_NUMERICSERV
+   IPPROTO_IPV6 IPV6_V6ONLY
+   NI_DGRAM NI_NUMERICHOST NI_NUMERICSERV NIx_NOHOST NIx_NOSERV
    SO_REUSEADDR SO_REUSEPORT SO_BROADCAST SO_ERROR
    SOCK_DGRAM SOCK_STREAM 
    SOL_SOCKET
@@ -27,7 +27,7 @@ use Socket 1.95 qw(
 my $AF_INET6 = eval { Socket::AF_INET6() }; # may not be defined
 my $AI_ADDRCONFIG = eval { Socket::AI_ADDRCONFIG() } || 0;
 use POSIX qw( dup2 );
-use Errno qw( EINPROGRESS );
+use Errno qw( EINVAL EINPROGRESS );
 
 use constant HAVE_MSWIN32 => ( $^O eq "MSWin32" );
 
@@ -135,6 +135,28 @@ sub import
    goto &IO::Socket::import;
 }
 
+# Convenient capability test function
+{
+   my $can_disable_v6only;
+   sub CAN_DISABLE_V6ONLY
+   {
+      return $can_disable_v6only if defined $can_disable_v6only;
+
+      socket my $testsock, Socket::PF_INET6, SOCK_STREAM, 0 or
+         die "Cannot socket(PF_INET6) - $!";
+
+      if( setsockopt $testsock, IPPROTO_IPV6, IPV6_V6ONLY, 0 ) {
+         return $can_disable_v6only = 1;
+      }
+      elsif( $! == EINVAL ) {
+         return $can_disable_v6only = 0;
+      }
+      else {
+         die "Cannot setsockopt() - $!";
+      }
+   }
+}
+
 =head1 CONSTRUCTORS
 
 =cut
@@ -239,6 +261,21 @@ If not defined, the socket option will not be changed, and default value set
 by the operating system will apply. For repeatable behaviour across platforms
 it is recommended this value always be defined for listening-mode sockets.
 
+Note that not all platforms support disabling this option. Some, at least
+OpenBSD and MirBSD, will fail with C<EINVAL> if you attempt to disable it.
+To determine whether it is possible to disable, you may use the class method
+
+ if( IO::Socket::IP->CAN_DISABLE_V6ONLY ) {
+    ...
+ }
+ else {
+    ...
+ }
+
+If your platform does not support disabling this option but you still want to
+listen for both C<AF_INET> and C<AF_INET6> connections you will have to create
+two listening sockets, one bound to each protocol.
+
 =item Timeout
 
 This C<IO::Socket::INET>-style argument is not currently supported. See the
@@ -328,8 +365,6 @@ sub _configure
    my @localinfos;
    my @peerinfos;
 
-   my @sockopts_enabled;
-
    $hints{flags} = $AI_ADDRCONFIG;
 
    if( defined $arg->{Family} ) {
@@ -413,6 +448,7 @@ sub _configure
       $err and ( $@ = "$err", return );
    }
 
+   my @sockopts_enabled;
    push @sockopts_enabled, SO_REUSEADDR if delete $arg->{ReuseAddr};
    push @sockopts_enabled, SO_REUSEPORT if delete $arg->{ReusePort};
    push @sockopts_enabled, SO_BROADCAST if delete $arg->{Broadcast};
@@ -494,7 +530,7 @@ sub setup
 
       if( defined ${*$self}{io_socket_ip_v6only} and defined $AF_INET6 and $info->{family} == $AF_INET6 ) {
          my $v6only = ${*$self}{io_socket_ip_v6only};
-         $self->setsockopt( Socket::IPPROTO_IPV6, IPV6_V6ONLY, pack "i", $v6only ) or ( $@ = "$!", return undef );
+         $self->setsockopt( IPPROTO_IPV6, IPV6_V6ONLY, pack "i", $v6only ) or ( $@ = "$!", return undef );
       }
 
       if( defined( my $addr = $info->{localaddr} ) ) {
@@ -559,14 +595,11 @@ L<IO::Socket> and L<IO::Handle>.
 sub _get_host_service
 {
    my $self = shift;
-   my ( $addr, $numeric ) = @_;
-
-   my $flags = 0;
+   my ( $addr, $flags, $xflags ) = @_;
 
    $flags |= NI_DGRAM if $self->socktype == SOCK_DGRAM;
-   $flags |= NI_NUMERICHOST|NI_NUMERICSERV if $numeric;
 
-   my ( $err, $host, $service ) = getnameinfo( $addr, $flags );
+   my ( $err, $host, $service ) = getnameinfo( $addr, $flags, $xflags || 0 );
    croak "getnameinfo - $err" if $err;
 
    return ( $host, $service );
@@ -592,7 +625,7 @@ sub sockhost_service
    my $self = shift;
    my ( $numeric ) = @_;
 
-   $self->_get_host_service( $self->sockname, $numeric );
+   $self->_get_host_service( $self->sockname, $numeric ? NI_NUMERICHOST|NI_NUMERICSERV : 0 );
 }
 
 =head2 $addr = $sock->sockhost
@@ -613,11 +646,11 @@ Return the resolved name of the local port number
 
 =cut
 
-sub sockhost { ( shift->sockhost_service(1) )[0] }
-sub sockport { ( shift->sockhost_service(1) )[1] }
+sub sockhost { my $self = shift; ( $self->_get_host_service( $self->sockname, NI_NUMERICHOST, NIx_NOSERV ) )[0] }
+sub sockport { my $self = shift; ( $self->_get_host_service( $self->sockname, NI_NUMERICSERV, NIx_NOHOST ) )[1] }
 
-sub sockhostname { ( shift->sockhost_service(0) )[0] }
-sub sockservice  { ( shift->sockhost_service(0) )[1] }
+sub sockhostname { my $self = shift; ( $self->_get_host_service( $self->sockname, 0, NIx_NOSERV ) )[0] }
+sub sockservice  { my $self = shift; ( $self->_get_host_service( $self->sockname, 0, NIx_NOHOST ) )[1] }
 
 =head2 ( $host, $service ) = $sock->peerhost_service( $numeric )
 
@@ -637,7 +670,7 @@ sub peerhost_service
    my $self = shift;
    my ( $numeric ) = @_;
 
-   $self->_get_host_service( $self->peername, $numeric );
+   $self->_get_host_service( $self->peername, $numeric ? NI_NUMERICHOST|NI_NUMERICSERV : 0 );
 }
 
 =head2 $addr = $sock->peerhost
@@ -658,11 +691,11 @@ Return the resolved name of the peer port number
 
 =cut
 
-sub peerhost    { ( shift->peerhost_service(1) )[0] }
-sub peerport    { ( shift->peerhost_service(1) )[1] }
+sub peerhost { my $self = shift; ( $self->_get_host_service( $self->peername, NI_NUMERICHOST, NIx_NOSERV ) )[0] }
+sub peerport { my $self = shift; ( $self->_get_host_service( $self->peername, NI_NUMERICSERV, NIx_NOHOST ) )[1] }
 
-sub peerhostname { ( shift->peerhost_service(0) )[0] }
-sub peerservice  { ( shift->peerhost_service(0) )[1] }
+sub peerhostname { my $self = shift; ( $self->_get_host_service( $self->peername, 0, NIx_NOSERV ) )[0] }
+sub peerservice  { my $self = shift; ( $self->_get_host_service( $self->peername, 0, NIx_NOHOST ) )[1] }
 
 # This unbelievably dodgy hack works around the bug that IO::Socket doesn't do
 # it
@@ -809,34 +842,11 @@ useable address from the results of the C<getaddrinfo(3)> call.
 
 =item *
 
-Cache the returns from C<sockhost_service> and C<peerhost_service> to avoid
-double-lookup overhead in such code as
-
-  printf "Peer is %s:%d\n", $sock->peerhost, $sock->peerport;
-
-=item *
-
 Investigate whether C<POSIX::dup2> upsets BSD's C<kqueue> watchers, and if so,
 consider what possible workarounds might be applied.
 
 =back
 
-=head1 BUGS
-
-=over 4
-
-=item *
-
-Nonblocking connect fails unit tests on MSWin32 smoke-testing machines. The
-specifics of the failure are that C<connect()> seems to block anyway despite
-being asked not to, and that failure to connect is not detected properly. I am
-as yet unsure why this is.
-
-Blocking connect on MSWin32, and both blocking and nonblocking connect on
-other platforms, all test OK on smoke testing.
-
-=back
-
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
